from collections.abc import Sequence
from typing import Self

from draive.lmm import AnyTool, Toolbox
from draive.parameters import State
from draive.types import MultimodalContent, MultimodalContentConvertible

__all__ = [
    "Step",
]


class Step(State):
    @classmethod
    def of(
        cls,
        input: MultimodalContent | MultimodalContentConvertible,  # noqa: A002
        /,
        *,
        tools: Toolbox | Sequence[AnyTool] | None = None,
    ) -> Self:
        toolbox: Toolbox
        match tools:
            case None:
                toolbox = Toolbox()

            case Toolbox() as tools:
                toolbox = tools

            case sequence:
                toolbox = Toolbox(*sequence)

        return cls(
            input=MultimodalContent.of(input),
            toolbox=toolbox,
        )

    input: MultimodalContent
    toolbox: Toolbox



from draive.steps.call import steps_completion
from draive.steps.completion import StepsCompletion
from draive.steps.lmm import lmm_steps_completion
from draive.steps.model import Step
from draive.steps.state import Steps

__all__ = [
    "lmm_steps_completion",
    "steps_completion",
    "Steps",
    "StepsCompletion",
    "Step",
]



from collections.abc import Sequence
from datetime import UTC, datetime
from typing import Literal, overload

from draive.conversation.model import ConversationMessage, ConversationResponseStream
from draive.conversation.state import Conversation
from draive.helpers import ConstantMemory
from draive.instructions import Instruction
from draive.lmm import AnyTool, Toolbox
from draive.scope import ctx
from draive.types import Memory, MultimodalContent, MultimodalContentConvertible

__all__ = [
    "conversation_completion",
]


@overload
async def conversation_completion(
    *,
    instruction: Instruction | str,
    input: ConversationMessage | MultimodalContent | MultimodalContentConvertible,
    memory: Memory[Sequence[ConversationMessage], ConversationMessage]
    | Sequence[ConversationMessage]
    | None = None,
    tools: Toolbox | Sequence[AnyTool] | None = None,
    stream: Literal[True],
) -> ConversationResponseStream: ...


@overload
async def conversation_completion(
    *,
    instruction: Instruction | str,
    input: ConversationMessage | MultimodalContent | MultimodalContentConvertible,
    memory: Memory[Sequence[ConversationMessage], ConversationMessage]
    | Sequence[ConversationMessage]
    | None = None,
    tools: Toolbox | Sequence[AnyTool] | None = None,
    stream: Literal[False] = False,
) -> ConversationMessage: ...


@overload
async def conversation_completion(
    *,
    instruction: Instruction | str,
    input: ConversationMessage | MultimodalContent | MultimodalContentConvertible,
    memory: Memory[Sequence[ConversationMessage], ConversationMessage]
    | Sequence[ConversationMessage]
    | None = None,
    tools: Toolbox | Sequence[AnyTool] | None = None,
    stream: bool,
) -> ConversationResponseStream | ConversationMessage: ...


async def conversation_completion(
    *,
    instruction: Instruction | str,
    input: ConversationMessage | MultimodalContent | MultimodalContentConvertible,  # noqa: A002
    memory: Memory[Sequence[ConversationMessage], ConversationMessage]
    | Sequence[ConversationMessage]
    | None = None,
    tools: Toolbox | Sequence[AnyTool] | None = None,
    stream: bool = False,
) -> ConversationResponseStream | ConversationMessage:
    conversation: Conversation = ctx.state(Conversation)

    # prepare message
    message: ConversationMessage
    match input:
        case ConversationMessage() as conversation_message:
            if guardrails := conversation.guardrails:
                message = conversation_message.updated(
                    content=await guardrails(conversation_message.content),
                )

            else:
                message = conversation_message

        case content:
            if guardrails := conversation.guardrails:
                message = ConversationMessage(
                    role="user",
                    created=datetime.now(UTC),
                    content=await guardrails(MultimodalContent.of(content)),
                )

            else:
                message = ConversationMessage(
                    role="user",
                    created=datetime.now(UTC),
                    content=MultimodalContent.of(content),
                )

    # prepare memory
    conversation_memory: Memory[Sequence[ConversationMessage], ConversationMessage]
    match memory or conversation.memory:
        case None:
            conversation_memory = ConstantMemory([])

        case Memory() as memory:
            conversation_memory = memory

        case [*memory_messages]:
            conversation_memory = ConstantMemory(memory_messages)

    # prepare tools
    toolbox: Toolbox
    match tools:
        case None:
            toolbox = Toolbox()

        case Toolbox() as tools:
            toolbox = tools

        case [*tools]:
            toolbox = Toolbox(*tools)

    # request completion
    return await conversation.completion(
        instruction=instruction,
        message=message,
        memory=conversation_memory,
        toolbox=toolbox,
        stream=stream,
    )



from typing import Any, Protocol, runtime_checkable

from draive.types import MultimodalContent

__all__ = [
    "ContentGuardrails",
]


@runtime_checkable
class ContentGuardrails(Protocol):
    async def __call__(
        self,
        content: MultimodalContent,
        /,
        **extra: Any,
    ) -> MultimodalContent: ...



from collections.abc import Iterable
from typing import Any

from draive.instructions import Instruction
from draive.scope import ctx
from draive.steps.model import Step
from draive.steps.state import Steps
from draive.types import Multimodal, MultimodalContent

__all__ = [
    "steps_completion",
]


async def steps_completion(
    *,
    instruction: Instruction | str,
    steps: Iterable[Step | Multimodal],
    **extra: Any,
) -> MultimodalContent:
    return await ctx.state(Steps).completion(
        instruction=instruction,
        steps=steps,
        **extra,
    )



from collections.abc import Iterable, Sequence
from typing import Any, Protocol, runtime_checkable

from draive.choice.model import ChoiceOption
from draive.instructions import Instruction
from draive.lmm import Toolbox
from draive.types import Multimodal

__all__ = [
    "ChoiceCompletion",
]


@runtime_checkable
class ChoiceCompletion(Protocol):
    async def __call__(
        self,
        *,
        instruction: Instruction | str,
        options: Sequence[ChoiceOption],
        input: Multimodal,  # noqa: A002
        toolbox: Toolbox,
        examples: Iterable[tuple[Multimodal, ChoiceOption]] | None,
        **extra: Any,
    ) -> ChoiceOption: ...



from draive.choice.call import choice_completion
from draive.choice.completion import ChoiceCompletion
from draive.choice.errors import SelectionException
from draive.choice.lmm import lmm_choice_completion
from draive.choice.model import ChoiceOption
from draive.choice.state import Choice

__all__ = [
    "choice_completion",
    "Choice",
    "ChoiceCompletion",
    "ChoiceOption",
    "lmm_choice_completion",
    "SelectionException",
]



from asyncio import Lock, gather
from collections.abc import Callable
from pathlib import Path
from typing import Protocol, Self, overload, runtime_checkable
from uuid import UUID, uuid4

from draive.evaluation.evaluator import EvaluatorResult, PreparedEvaluator
from draive.evaluation.scenario import PreparedScenarioEvaluator, ScenarioEvaluatorResult
from draive.parameters import DataModel, Field
from draive.scope import ctx
from draive.types import frozenlist
from draive.utils import asynchronous

__all__ = [
    "evaluation_suite",
    "EvaluationCaseResult",
    "EvaluationSuite",
    "EvaluationSuiteCaseResult",
    "EvaluationSuiteDefinition",
    "EvaluationSuiteStorage",
]


class EvaluationSuiteCase[CaseParameters: DataModel](DataModel):
    identifier: UUID = Field(default_factory=uuid4)
    parameters: CaseParameters
    comment: str | None = None


class EvaluationSuiteCaseResult[CaseParameters: DataModel, Value: DataModel | str](DataModel):
    case: EvaluationSuiteCase[CaseParameters] = Field(
        description="Evaluated case",
    )
    value: Value = Field(
        description="Evaluated value",
    )
    results: frozenlist[ScenarioEvaluatorResult | EvaluatorResult] = Field(
        description="Evaluation results",
    )

    @property
    def passed(self) -> bool:
        return all(result.passed for result in self.results)


class EvaluationCaseResult[Value: DataModel | str](DataModel):
    @classmethod
    def of(
        cls,
        results: ScenarioEvaluatorResult | EvaluatorResult,
        *_results: ScenarioEvaluatorResult | EvaluatorResult,
        value: Value,
    ) -> Self:
        return cls(
            value=value,
            results=(results, *_results),
        )

    @classmethod
    async def evaluating(
        cls,
        value: Value,
        /,
        evaluators: PreparedScenarioEvaluator[Value] | PreparedEvaluator[Value],
        *_evaluators: PreparedScenarioEvaluator[Value] | PreparedEvaluator[Value],
    ) -> Self:
        return cls(
            value=value,
            results=tuple(
                await gather(
                    *[evaluator(value) for evaluator in [evaluators, *_evaluators]],
                    return_exceptions=False,
                ),
            ),
        )

    value: Value = Field(
        description="Evaluated value",
    )
    results: frozenlist[ScenarioEvaluatorResult | EvaluatorResult] = Field(
        description="Evaluation results",
    )


@runtime_checkable
class EvaluationSuiteDefinition[CaseParameters: DataModel, Value: DataModel | str](Protocol):
    async def __call__(
        self,
        parameters: CaseParameters,
    ) -> EvaluationCaseResult[Value]: ...


class EvaluationSuiteData[CaseParameters: DataModel](DataModel):
    cases: frozenlist[EvaluationSuiteCase[CaseParameters]] = Field(default_factory=tuple)


@runtime_checkable
class EvaluationSuiteStorage[CaseParameters: DataModel](Protocol):
    async def load(
        self,
    ) -> EvaluationSuiteData[CaseParameters]: ...

    async def save(
        self,
        data: EvaluationSuiteData[CaseParameters],
    ) -> None: ...


class EvaluationSuite[CaseParameters: DataModel, Value: DataModel | str]:
    def __init__(
        self,
        definition: EvaluationSuiteDefinition[CaseParameters, Value],
        storage: EvaluationSuiteStorage[CaseParameters],
    ) -> None:
        self._definition: EvaluationSuiteDefinition[CaseParameters, Value] = definition
        self._storage: EvaluationSuiteStorage[CaseParameters] = storage
        self._data_cache: EvaluationSuiteData[CaseParameters] | None = None
        self._lock: Lock = Lock()

    @overload
    async def __call__(
        self,
        parameters: CaseParameters | UUID | None,
        /,
        *,
        reload: bool = False,
    ) -> EvaluationSuiteCaseResult[CaseParameters, Value]: ...

    @overload
    async def __call__(
        self,
        /,
        *,
        reload: bool = False,
    ) -> list[EvaluationSuiteCaseResult[CaseParameters, Value]]: ...

    async def __call__(
        self,
        parameters: CaseParameters | UUID | None = None,
        /,
        *,
        reload: bool = False,
    ) -> (
        list[EvaluationSuiteCaseResult[CaseParameters, Value]]
        | EvaluationSuiteCaseResult[CaseParameters, Value]
    ):
        async with self._lock:
            match parameters:
                case None:
                    return await gather(
                        *[
                            self._evaluate(case=case)
                            for case in (await self._data(reload=reload)).cases
                        ],
                        return_exceptions=False,
                    )

                case UUID() as identifier:
                    available_cases: frozenlist[EvaluationSuiteCase[CaseParameters]] = (
                        await self._data(reload=reload)
                    ).cases

                    if evaluation_case := next(
                        iter([case for case in available_cases if case.identifier == identifier]),
                        None,
                    ):
                        return await self._evaluate(case=evaluation_case)

                    else:
                        raise ValueError(f"Evaluation case with ID {identifier} does not exists.")

                case case_parameters:
                    return await self._evaluate(
                        case=EvaluationSuiteCase[CaseParameters](
                            parameters=case_parameters,
                        )
                    )

    async def _evaluate(
        self,
        *,
        case: EvaluationSuiteCase[CaseParameters],
    ) -> EvaluationSuiteCaseResult[CaseParameters, Value]:
        result: EvaluationCaseResult[Value] = await self._definition(parameters=case.parameters)

        return EvaluationSuiteCaseResult[CaseParameters, Value](
            case=case,
            value=result.value,
            results=result.results,
        )

    async def _data(
        self,
        reload: bool = False,
    ) -> EvaluationSuiteData[CaseParameters]:
        if reload or self._data_cache is None:
            self._data_cache = await self._storage.load()
            return self._data_cache

        else:
            return self._data_cache

    async def cases(
        self,
        reload: bool = False,
    ) -> frozenlist[EvaluationSuiteCase[CaseParameters]]:
        async with self._lock:
            return (await self._data(reload=reload)).cases

    async def add_case(
        self,
        parameters: CaseParameters,
        /,
        comment: str | None = None,
    ) -> None:
        async with self._lock:
            data: EvaluationSuiteData[CaseParameters] = await self._data(reload=True)
            self._data_cache = data.updated(
                cases=(
                    *data.cases,
                    EvaluationSuiteCase[CaseParameters](
                        parameters=parameters,
                        comment=comment,
                    ),
                )
            )
            await self._storage.save(self._data_cache)

    async def remove_case(
        self,
        identifier: UUID,
        /,
    ) -> None:
        async with self._lock:
            data: EvaluationSuiteData[CaseParameters] = await self._data(reload=True)
            self._data_cache = data.updated(
                cases=tuple(case for case in data.cases if case.identifier != identifier)
            )
            await self._storage.save(self._data_cache)


def evaluation_suite[CaseParameters: DataModel, Value: DataModel | str](
    case: type[CaseParameters],
    /,
    storage: EvaluationSuiteStorage[CaseParameters] | Path | str | None = None,
) -> Callable[
    [EvaluationSuiteDefinition[CaseParameters, Value]],
    EvaluationSuite[CaseParameters, Value],
]:
    suite_storage: EvaluationSuiteStorage[CaseParameters]
    match storage:
        case None:
            suite_storage = _EvaluationSuiteMemoryStorage[CaseParameters](
                data_type=EvaluationSuiteData[case],
            )

        case str() as path_str:
            suite_storage = _EvaluationSuiteFileStorage[CaseParameters](
                path=path_str,
                data_type=EvaluationSuiteData[case],
            )

        case Path() as path:
            suite_storage = _EvaluationSuiteFileStorage[CaseParameters](
                path=path,
                data_type=EvaluationSuiteData[case],
            )

        case storage:
            suite_storage = storage

    def wrap(
        definition: EvaluationSuiteDefinition[CaseParameters, Value],
    ) -> EvaluationSuite[CaseParameters, Value]:
        return EvaluationSuite[CaseParameters, Value](
            definition=definition,
            storage=suite_storage,
        )

    return wrap


class _EvaluationSuiteMemoryStorage[CaseParameters: DataModel]:
    def __init__(
        self,
        data_type: type[EvaluationSuiteData[CaseParameters]],
    ) -> None:
        self.store: frozenlist[EvaluationSuiteCase[CaseParameters]] = ()
        self._data_type: type[EvaluationSuiteData[CaseParameters]] = data_type

    async def load(
        self,
    ) -> EvaluationSuiteData[CaseParameters]:
        return self._data_type(cases=self.store)

    async def save(
        self,
        data: EvaluationSuiteData[CaseParameters],
    ) -> None:
        self.store = data.cases


class _EvaluationSuiteFileStorage[CaseParameters: DataModel]:
    def __init__(
        self,
        path: Path | str,
        data_type: type[EvaluationSuiteData[CaseParameters]],
    ) -> None:
        self._path: Path
        match path:
            case str() as path_str:
                self._path = Path(path_str)

            case path:
                self._path = path

        self._data_type: type[EvaluationSuiteData[CaseParameters]] = data_type

    async def load(
        self,
    ) -> EvaluationSuiteData[CaseParameters]:
        try:
            return await self._file_load(data_type=self._data_type)

        except ValueError as exc:
            ctx.log_error(
                f"Invalid EvaluationSuite at {self._path}",
                exception=exc,
            )
            return self._data_type()

    @asynchronous(executor=None)
    def _file_load(
        self,
        data_type: type[EvaluationSuiteData[CaseParameters]],
    ) -> EvaluationSuiteData[CaseParameters]:
        if self._path.exists():
            with open(self._path, mode="rb") as file:
                return data_type.from_json(file.read())

        else:
            return data_type()

    async def save(
        self,
        data: EvaluationSuiteData[CaseParameters],
    ) -> None:
        await self._file_save(data=data)

    @asynchronous(executor=None)
    def _file_save(
        self,
        data: EvaluationSuiteData[CaseParameters],
    ) -> None:
        with open(self._path, mode="wb+") as file:
            file.write(data.as_json(indent=2).encode("utf-8"))



from asyncio import gather
from typing import Any

from openai.types.moderation import Moderation

from draive.openai.client import OpenAIClient
from draive.openai.config import OpenAIModerationConfig
from draive.safeguards import GuardrailsException
from draive.scope import ctx
from draive.splitters import split_text
from draive.types import MultimodalContent
from draive.utils.missing import not_missing

__all__ = [
    "openai_content_guardrails",
]


async def openai_content_guardrails(  # noqa: C901, PLR0912, PLR0915
    content: MultimodalContent,
    /,
    **extra: Any,
) -> MultimodalContent:
    ctx.log_debug("Requested OpenAI content guardrails")
    client: OpenAIClient = ctx.dependency(OpenAIClient)
    config: OpenAIModerationConfig = ctx.state(OpenAIModerationConfig).updated(**extra)

    content_parts: list[str] = split_text(
        content.as_string(),
        part_size=2000,
        part_overlap_size=200,
        count_size=len,
    )
    results: list[Moderation] = await gather(
        *[client.moderation_check(text=part) for part in content_parts]
    )

    violations: set[str] = set()
    for result in results:
        if (
            not_missing(config.harassment_threshold)
            and result.category_scores.harassment >= config.harassment_threshold
        ):
            violations.add("harassment")

        elif result.categories.harassment:
            violations.add("harassment")

        if (
            not_missing(config.harassment_threatening_threshold)
            and result.category_scores.harassment_threatening
            >= config.harassment_threatening_threshold
        ):
            violations.add("harassment_threatening")

        elif result.categories.harassment_threatening:
            violations.add("harassment_threatening")

        if (
            not_missing(config.hate_threshold)
            and result.category_scores.hate >= config.hate_threshold
        ):
            violations.add("hate")

        elif result.categories.hate:
            violations.add("hate")

        if (
            not_missing(config.hate_threatening_threshold)
            and result.category_scores.hate_threatening >= config.hate_threatening_threshold
        ):
            violations.add("hate_threatening")

        elif result.categories.hate_threatening:
            violations.add("hate_threatening")

        if (
            not_missing(config.self_harm_threshold)
            and result.category_scores.self_harm >= config.self_harm_threshold
        ):
            violations.add("self_harm")

        elif result.categories.self_harm:
            violations.add("self_harm")

        if (
            not_missing(config.self_harm_instructions_threshold)
            and result.category_scores.self_harm_instructions
            >= config.self_harm_instructions_threshold
        ):
            violations.add("self_harm_instructions")

        elif result.categories.self_harm_instructions:
            violations.add("self_harm_instructions")

        if (
            not_missing(config.self_harm_intent_threshold)
            and result.category_scores.self_harm_intent >= config.self_harm_intent_threshold
        ):
            violations.add("self_harm_intent")

        elif result.categories.self_harm_intent:
            violations.add("self_harm_intent")

        if (
            not_missing(config.sexual_threshold)
            and result.category_scores.sexual >= config.sexual_threshold
        ):
            violations.add("sexual")

        elif result.categories.sexual:
            violations.add("sexual")

        if (
            not_missing(config.sexual_minors_threshold)
            and result.category_scores.sexual_minors >= config.sexual_minors_threshold
        ):
            violations.add("sexual_minors")

        elif result.categories.sexual_minors:
            violations.add("sexual_minors")

        if (
            not_missing(config.violence_threshold)
            and result.category_scores.violence >= config.violence_threshold
        ):
            violations.add("violence")

        elif result.categories.violence:
            violations.add("violence")

        if (
            not_missing(config.violence_graphic_threshold)
            and result.category_scores.violence_graphic >= config.violence_graphic_threshold
        ):
            violations.add("violence_graphic")

        elif result.categories.violence_graphic:
            violations.add("violence_graphic")

    if violations:
        raise GuardrailsException(f"Content violated rule(s): {violations}")

    return content  # no violation - use content



from collections.abc import Iterable, Sequence
from typing import Any

from draive.choice.model import ChoiceOption
from draive.choice.state import Choice
from draive.instructions import Instruction
from draive.lmm import AnyTool, Toolbox
from draive.scope import ctx
from draive.types import Multimodal

__all__ = [
    "choice_completion",
]


async def choice_completion(
    *,
    instruction: Instruction | str,
    options: Iterable[ChoiceOption | Multimodal],
    input: Multimodal,  # noqa: A002
    tools: Toolbox | Sequence[AnyTool] | None = None,
    examples: Iterable[tuple[Multimodal, ChoiceOption]] | None = None,
    **extra: Any,
) -> ChoiceOption:
    toolbox: Toolbox
    match tools:
        case None:
            toolbox = Toolbox()

        case Toolbox() as tools:
            toolbox = tools

        case sequence:
            toolbox = Toolbox(*sequence)

    return await ctx.state(Choice).completion(
        instruction=instruction,
        options=[
            option if isinstance(option, ChoiceOption) else ChoiceOption.of(option)
            for option in options
        ],
        input=input,
        toolbox=toolbox,
        examples=examples,
        **extra,
    )



from draive.choice.completion import ChoiceCompletion
from draive.choice.lmm import lmm_choice_completion
from draive.parameters import State

__all__: list[str] = [
    "Choice",
]


class Choice(State):
    completion: ChoiceCompletion = lmm_choice_completion



from typing import Any, Self
from uuid import uuid4

from draive.parameters import State
from draive.types import Multimodal, MultimodalContent

__all__ = [
    "ChoiceOption",
]


class ChoiceOption(State):
    @classmethod
    def of(
        cls,
        content: Multimodal,
        /,
        identifier: str | None = None,
        **meta: Any,
    ) -> Self:
        assert identifier is None or len(identifier) > 0, "Identifier can't be empty"  # nosec: B101

        return cls(
            identifier=identifier or uuid4().hex,
            content=MultimodalContent.of(content),
            meta=meta,
        )

    identifier: str
    content: MultimodalContent
    meta: dict[str, Any]



from draive.parameters import State
from draive.steps.completion import StepsCompletion
from draive.steps.lmm import lmm_steps_completion

__all__: list[str] = [
    "Steps",
]


class Steps(State):
    completion: StepsCompletion = lmm_steps_completion



from collections.abc import Iterable
from typing import Any

from draive.instructions import Instruction
from draive.lmm import lmm_invocation
from draive.lmm.tools.toolbox import Toolbox
from draive.scope import ctx
from draive.steps.model import Step
from draive.types import (
    LMMContextElement,
    LMMInput,
    Multimodal,
    MultimodalContent,
)
from draive.types.lmm import LMMCompletion, LMMToolRequests, LMMToolResponse

__all__ = [
    "lmm_steps_completion",
]


async def lmm_steps_completion(
    *,
    instruction: Instruction | str,
    steps: Iterable[Step | Multimodal],
    **extra: Any,
) -> MultimodalContent:
    with ctx.nested(
        "lmm_steps_completion",
    ):
        assert steps, "Steps cannot be empty"  # nosec: B101

        context: list[LMMContextElement] = []
        for step in steps:
            current_step: Step
            match step:
                case Step() as step:
                    current_step = step

                case input_content:
                    current_step = Step.of(input_content)

            await _lmm_process_step(
                current_step,
                instruction=instruction,
                context=context,
                **extra,
            )

        match context[-1]:  # the last element of the context should be the result
            case LMMCompletion() as result:
                return result.content

            case _:
                raise RuntimeError("Invalid steps completion state!")


async def _lmm_process_step(
    step: Step,
    /,
    instruction: Instruction | str,
    context: list[LMMContextElement],
    **extra: Any,
) -> None:
    context.append(LMMInput.of(step.input))
    toolbox: Toolbox = step.toolbox

    recursion_level: int = 0
    while recursion_level <= toolbox.recursion_limit:
        match await lmm_invocation(
            instruction=instruction,
            context=context,
            tools=toolbox.available_tools(),
            tool_selection=toolbox.tool_selection(recursion_level=recursion_level),
            output="text",
            stream=False,
            **extra,
        ):
            case LMMCompletion() as completion:
                ctx.log_debug("Received step result")
                return context.append(completion)

            case LMMToolRequests() as tool_requests:
                ctx.log_debug("Received step tool calls")
                responses: list[LMMToolResponse] = await toolbox.respond(tool_requests)

                if direct_content := [
                    response.content for response in responses if response.direct
                ]:
                    return context.append(LMMCompletion.of(MultimodalContent.of(*direct_content)))

                else:
                    context.extend([tool_requests, *responses])

        recursion_level += 1  # continue with next recursion level

    raise RuntimeError("LMM exceeded limit of recursive calls")



from draive.parameters import DataModel

__all__ = [
    "AudioBase64Content",
    "AudioContent",
    "AudioURLContent",
]


class AudioURLContent(DataModel):
    mime_type: str | None = None
    audio_url: str
    audio_transcription: str | None = None
    meta: dict[str, str | float | int | bool | None] | None = None

    def __bool__(self) -> bool:
        return bool(self.audio_url)


class AudioBase64Content(DataModel):
    mime_type: str | None = None
    audio_base64: str
    audio_transcription: str | None = None
    meta: dict[str, str | float | int | bool | None] | None = None

    def __bool__(self) -> bool:
        return bool(self.audio_base64)


AudioContent = AudioURLContent | AudioBase64Content



from collections.abc import Iterable
from typing import Any, Protocol, runtime_checkable

from draive.instructions import Instruction
from draive.steps.model import Step
from draive.types import Multimodal, MultimodalContent

__all__ = [
    "StepsCompletion",
]


@runtime_checkable
class StepsCompletion(Protocol):
    async def __call__(
        self,
        *,
        instruction: Instruction | str,
        steps: Iterable[Step | Multimodal],
        **extra: Any,
    ) -> MultimodalContent: ...
